@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 
 # Hive Appier Framework
-# Copyright (c) 2008-2015 Hive Solutions Lda.
+# Copyright (c) 2008-2016 Hive Solutions Lda.
 #
 # This file is part of Hive Appier Framework.
 #
@@ -31,13 +31,14 @@ __revision__ = "$LastChangedRevision$"
 __date__ = "$LastChangedDate$"
 """ The last change date of the module """
 
-__copyright__ = "Copyright (c) 2008-2015 Hive Solutions Lda."
+__copyright__ = "Copyright (c) 2008-2016 Hive Solutions Lda."
 """ The copyright for the module """
 
 __license__ = "Apache License, Version 2.0"
 """ The license for the module """
 
 import os
+import json
 import time
 import datetime
 import tempfile
@@ -75,6 +76,14 @@ class AdminPart(
     def load(self):
         appier.Part.load(self)
 
+        self.logger.debug("Registering pre request handler ...")
+        appier.App.add_custom("before_request", self.before_request)
+        appier.App.add_exception(
+            BaseException,
+            self.exception_handler,
+            scope = AdminPart
+        )
+
         self.logger.debug("Updating pre-defined application routes ...")
         self.owner.login_route = "admin.login"
         self.owner.login_route_admin = "admin.login"
@@ -87,6 +96,7 @@ class AdminPart(
         self.owner.admin_google_scope = ("email",)
         self.owner.admin_live_scope = ("wl.basic", "wl.emails")
 
+        self.owner.lib_loaders["appier_extras"] = self._appier_extras_loader
         self.owner.lib_loaders["netius"] = self._netius_loader
         self.owner.lib_loaders["PIL"] = self._pil_loader
         self.owner.lib_loaders["pymongo"] = self._pymongo_loader
@@ -95,6 +105,7 @@ class AdminPart(
 
         self.logger.debug("Generating admin interfaces ...")
         for model_c in self.models_r:
+            if not model_c.is_attached(): continue
             self.logger.debug(model_c)
 
     def routes(self):
@@ -105,19 +116,31 @@ class AdminPart(
             (("GET", "POST"), "/admin/signout", self.logout),
             (("GET"), "/admin/recover", self.recover),
             (("POST"), "/admin/recover", self.recover_do),
+            (("GET"), "/admin/reset", self.reset),
+            (("POST"), "/admin/reset", self.reset_do),
             (("GET",), "/admin/options", self.options),
             (("POST",), "/admin/options", self.options_action),
             (("GET",), "/admin/status", self.status),
             (("GET",), "/admin/social", self.social),
+            (("GET",), "/admin/operations", self.operations),
             (("GET",), "/admin/routes", self.list_routes),
             (("GET",), "/admin/configs", self.list_configs),
             (("GET",), "/admin/libraries", self.list_libraries),
+            (("GET",), "/admin/operations/build_index", self.build_index),
+            (("GET",), "/admin/operations/test_email", self.test_email),
+            (("GET",), "/admin/operations/test_event", self.test_event),
             (("GET",), "/admin/sessions", self.list_sessions),
+            (("GET",), "/admin/sessions/empty", self.empty_sessions),
+            (("GET",), "/admin/sessions/<str:sid>", self.show_session),
+            (("GET",), "/admin/sessions/<str:sid>/delete", self.delete_session),
+            (("GET",), "/admin/counters", self.list_counters),
+            (("GET",), "/admin/events.csv", self.list_events_csv),
             (("GET",), "/admin/database", self.database),
             (("GET",), "/admin/database/export", self.database_export),
             (("GET",), "/admin/database/import", self.database_import),
             (("POST",), "/admin/database/import", self.database_import_do),
             (("GET",), "/admin/database/reset", self.database_reset),
+            (("GET",), "/admin/search", self.search),
             (("GET",), "/admin/accounts/new", self.new_account),
             (("POST",), "/admin/accounts", self.create_account),
             (("GET",), "/admin/accounts/<str:username>", self.show_account),
@@ -161,7 +184,8 @@ class AdminPart(
     def models(self):
         return models
 
-    def template(self, template, layout = "fluid", *args, **kwargs):
+    def template(self, template, layout = None, *args, **kwargs):
+        layout = self.session.get("layout", self.layout)
         template = "%s/%s" % (layout, template)
         return appier.Part.template(
             self,
@@ -172,6 +196,26 @@ class AdminPart(
             **kwargs
         )
 
+    def before_request(self):
+        key = self.get_field("skey", None)
+        key = self.get_field("secret_key", key)
+        key = self.request.get_header("X-Secret-Key", key)
+        if not key: return
+        try: account = models.Account.login_key(key)
+        except appier.OperationalError: pass
+        else: account._set_session(method = "set_t")
+
+    def exception_handler(self, error):
+        import traceback
+        lines = traceback.format_exc().splitlines()
+        lines = self._lines(lines)
+        lines = lines if self.owner.is_devel() else []
+        return self.template(
+            "error.html.tpl",
+            error = error,
+            lines = lines
+        )
+
     def index(self):
         return self.list_models()
 
@@ -204,10 +248,7 @@ class AdminPart(
 
         # updates the current session with the proper
         # values to correctly authenticate the user
-        self.session["username"] = account.username
-        self.session["email"] = account.email
-        self.session["type"] = account.type_s()
-        self.session["tokens"] = account.tokens()
+        account._set_session()
 
         # redirects the current operation to the next url or in
         # alternative to the root index of the administration
@@ -223,17 +264,7 @@ class AdminPart(
         # verifies the existence of the various account related session
         # attributes and in case they exist removes them from session as
         # the user is currently logging out from session
-        if "username" in self.session: del self.session["username"]
-        if "email" in self.session: del self.session["email"]
-        if "type" in self.session: del self.session["type"]
-        if "tokens" in self.session: del self.session["tokens"]
-        if "fb.access_token" in self.session: del self.session["fb.access_token"]
-        if "tw.oauth_token" in self.session: del self.session["tw.oauth_token"]
-        if "tw.oauth_token_secret" in self.session: del self.session["tw.oauth_token_secret"]
-        if "tw.oauth_temporary" in self.session: del self.session["tw.oauth_temporary"]
-        if "gg.access_token" in self.session: del self.session["gg.access_token"]
-        if "gh.access_token" in self.session: del self.session["gh.access_token"]
-        if "live.access_token" in self.session: del self.session["live.access_token"]
+        models.Account._unset_session()
 
         # runs the proper redirect operation, taking into account if the
         # next value has been provided or not
@@ -256,6 +287,44 @@ class AdminPart(
 
         return self.template("done.html.tpl")
 
+    def reset(self):
+        next = self.field("next")
+        reset_token = self.field(
+            "reset_token",
+            mandatory = True,
+            not_empty = True
+        )
+        self.account_c.validate_token(reset_token)
+        return self.template(
+            "reset.html.tpl",
+            next = next,
+            reset_token = reset_token
+        )
+
+    def reset_do(self):
+        next = self.field("next")
+        reset_token = self.field(
+            "reset_token",
+            mandatory = True,
+            not_empty = True
+        )
+        password = self.field("password")
+        password_confirm = self.field("password_confirm")
+        try: self.account_c.reset(reset_token, password, password_confirm)
+        except appier.AppierException as error:
+            return self.template(
+                "reset.html.tpl",
+                next = next,
+                reset_token = reset_token,
+                password = password,
+                password_confirm = password_confirm,
+                error = error.message
+            )
+
+        return self.redirect(
+            next or self.url_for(self.login_route_admin)
+        )
+
     def new_account(self):
         if not self.owner.admin_open: raise appier.SecurityError(
             message = "signup not allowed"
@@ -357,6 +426,13 @@ class AdminPart(
             linked = linked
         )
 
+    @appier.ensure(token = "admin")
+    def operations(self):
+        return self.template(
+            "operations.html.tpl",
+            section = "operations"
+        )
+
     @appier.ensure(token = "admin")
     def list_routes(self):
         return self.template(
@@ -386,6 +462,50 @@ class AdminPart(
             libraries = libraries
         )
 
+    @appier.ensure(token = "admin")
+    def build_index(self):
+        empty = self.field("empty", True, cast = bool)
+        if empty: models.Search.delete_c()
+        for model in self._attached(self.models_r):
+            model.build_index_g()
+        return self.redirect(
+            self.url_for(
+                "admin.operations",
+                message = "Search index built with success"
+            )
+        )
+
+    @appier.ensure(token = "admin")
+    def test_email(self):
+        receiver = appier.conf("TEST_EMAIL", None)
+        if not receiver: raise appier.OperationalError(
+            message = "No test email defined"
+        )
+        models.Base.send_email_g(
+            appier.get_app(),
+            "admin/email/test.html.tpl",
+            receivers = [receiver],
+            subject = "Test email",
+            title = "Test email"
+        )
+        return self.redirect(
+            self.url_for(
+                "admin.operations",
+                message = "Test email sent"
+            )
+        )
+
+    @appier.ensure(token = "admin")
+    def test_event(self):
+        name = appier.conf("TEST_EVENT", "test")
+        models.Event.notify_g(name)
+        return self.redirect(
+            self.url_for(
+                "admin.operations",
+                message = "Test event triggered"
+            )
+        )
+
     @appier.ensure(token = "admin")
     def list_sessions(self):
         return self.template(
@@ -394,6 +514,72 @@ class AdminPart(
             sessions = self.request.session_c.all()
         )
 
+    @appier.ensure(token = "admin")
+    def empty_sessions(self):
+        self.request.session_c.empty()
+        return self.redirect(
+            self.url_for(
+                "admin.list_sessions",
+                message = "Sessions emptied with success"
+            )
+        )
+
+    @appier.ensure(token = "admin")
+    def show_session(self, sid):
+        sid = str(sid)
+        return self.template(
+            "session.html.tpl",
+            section = "status",
+            session_s = self.request.session_c.get_s(sid)
+        )
+
+    @appier.ensure(token = "admin")
+    def delete_session(self, sid):
+        sid = str(sid)
+        self.request.session_c.expire(sid)
+        return self.redirect(
+            self.url_for(
+                "admin.list_sessions",
+                message = "Session deleted with success"
+            )
+        )
+
+    @appier.ensure(token = "admin")
+    def list_counters(self):
+        collection = self._counters()
+        counters = collection.find()
+        return self.template(
+            "counters.html.tpl",
+            section = "status",
+            counters = counters
+        )
+
+    @appier.ensure(token = "admin")
+    def list_events_csv(self):
+        object = appier.get_object(
+            alias = True,
+            find = True,
+            limit = 0
+        )
+        events = models.Event.find(**object)
+
+        events_s = [(
+            "name",
+            "handler",
+            "arguments"
+        )]
+        for event in events:
+            event_s = (
+                event.name,
+                event.handler,
+                event.arguments and json.dumps(event.arguments)
+            )
+            events_s.append(event_s)
+
+        result = appier.serialize_csv(events_s, delimiter = ",")
+        self.content_type("text/csv")
+        return result
+
     @appier.ensure(token = "admin")
     def database(self):
         return self.template(
@@ -403,11 +589,10 @@ class AdminPart(
 
     @appier.ensure(token = "admin")
     def database_export(self):
-        adapter = appier.get_adapter()
-        database = adapter.get_db()
+        adapter = self.get_adapter()
         file = appier.legacy.BytesIO()
         manager = appier.ExportManager(
-            database,
+            adapter,
             multiple = self.resolve()
         )
         manager.export_data(file)
@@ -451,10 +636,9 @@ class AdminPart(
         # retrieves the database and creates a new export manager for
         # the currently defined entities then imports the data defined
         # in the current temporary path
-        adapter = appier.get_adapter()
-        database = adapter.get_db()
+        adapter = self.get_adapter()
         manager = appier.ExportManager(
-            database,
+            adapter,
             multiple = self.resolve()
         )
         try: manager.import_data(file_path)
@@ -466,9 +650,20 @@ class AdminPart(
             )
         )
 
+    @appier.ensure(token = "admin")
+    def search(self):
+        object = appier.get_object(
+            alias = True,
+            find = True,
+            find_i = True,
+            find_t = "right"
+        )
+        indexes = models.Search.find(map = True, **object)
+        return indexes
+
     @appier.ensure(token = "admin")
     def database_reset(self):
-        adapter = appier.get_adapter()
+        adapter = self.get_adapter()
         adapter.drop_db()
         return self.redirect(
             self.url_for(
@@ -524,7 +719,7 @@ class AdminPart(
         parameters = self.get_fields("parameters", [])
         ids = self.field("ids", "")
         ids = ids.split(",")
-        ids = [appier.object_id(_id) for _id in ids if _id]
+        ids = [self.get_adapter().object_id(_id) for _id in ids if _id]
         is_global = self.field("is_global", False, cast = bool)
         model = self.get_model(model)
         if ids: kwargs = dict(_id = {"$in" : ids})
@@ -539,22 +734,70 @@ class AdminPart(
 
     @appier.ensure(token = "admin")
     def operation_model(self, model, operation):
+        # retrieves the complete set of fields that are going to
+        # be used while performing the operation and runs the
+        # complete set of casting operations for each of them
         parameters = self.get_fields("parameters", [])
         next = self.field("next")
         ids = self.field("ids", "")
         ids = ids.split(",")
-        ids = [appier.object_id(_id) for _id in ids if _id]
+        ids = [self.get_adapter().object_id(_id) for _id in ids if _id]
         is_global = self.field("is_global", False, cast = bool)
+
+        # retrieves the model information (class) for the provided
+        # name and then also the operation definition (metadata)
         model = self.get_model(model)
         definition = model.operation(operation)
+
+        # using the retrieved information determines if the operation
+        # is considered to be a factory one (generates entity) and
+        # then casts the provided parameters according to metadata
+        factory = definition.get("factory", False)
         parameters = definition.cast(parameters)
+
+        # determines the filter information that is going to be passed
+        # to the find operations and retrieves the associated entities
+        # that are going to be the target for the operations, note that
+        # if this is a global operation classes are used instead
         if ids: kwargs = dict(_id = {"$in" : ids})
         else: kwargs = dict()
         if is_global: entities = (model,)
         else: entities = model.find(**kwargs)
+
+        # determines if this is going to be a single entity target
+        # operation or if otherwise it's a multiple target one, taking
+        # that into account determines if this is considered a factory
+        # operation, note that only single operations are allowed
+        is_single = len(entities) == 1
+        factory = factory and is_single
+
+        # defines the original/default result value for the complete
+        # set of operations to be performed
+        result = None
+
+        # iterates over the complete set of selected entities and
+        # performs the requested operation for each of them
         for entity in entities:
             method = getattr(entity, operation)
-            method(*parameters)
+            result = method(*parameters)
+
+        # in case the factory mode is enabled, a new entity has been
+        # created and proper redirection must be performed so that the
+        # new entity is shown instead of the default show model
+        if factory:
+            cls = result.__class__
+            model_name = cls._name()
+            model_id = result._id
+            return self.redirect(
+                self.url_for(
+                    "admin.show_entity",
+                    model = model_name,
+                    _id = model_id
+                )
+            )
+
+        # runs the default redirection process that displays the model
+        # page for the model used as target for the operation
         return self.redirect(
             next or self.url_for("admin.show_model", model = model._name())
         )
@@ -594,7 +837,7 @@ class AdminPart(
         entity = model.get(
             rules = False,
             meta = True,
-            _id = appier.object_id(_id)
+            _id = self.get_adapter().object_id(_id)
         )
         return self.template(
             "entities/show.html.tpl",
@@ -612,7 +855,7 @@ class AdminPart(
             rules = rules,
             meta = meta,
             map = True,
-            _id = appier.object_id(_id)
+            _id = self.get_adapter().object_id(_id)
         )
         return entity
 
@@ -622,7 +865,7 @@ class AdminPart(
         entity = model.get(
             rules = False,
             meta = True,
-            _id = appier.object_id(_id)
+            _id = self.get_adapter().object_id(_id)
         )
         return self.template(
             "entities/edit.html.tpl",
@@ -638,7 +881,7 @@ class AdminPart(
         entity = model.get(
             rules = False,
             meta = True,
-            _id = appier.object_id(_id)
+            _id = self.get_adapter().object_id(_id)
         )
         entity.apply(safe_a = False)
         try: entity.save()
@@ -662,7 +905,7 @@ class AdminPart(
     @appier.ensure(token = "admin")
     def delete_entity(self, model, _id):
         model = self.get_model(model)
-        entity = model.get(_id = appier.object_id(_id))
+        entity = model.get(_id = self.get_adapter().object_id(_id))
         entity.delete()
         return self.redirect(
             self.url_for(
@@ -948,10 +1191,7 @@ class AdminPart(
 
         # updates the current session with the proper
         # values to correctly authenticate the user
-        self.session["username"] = account.username
-        self.session["email"] = account.email
-        self.session["type"] = account.type_s()
-        self.session["tokens"] = account.tokens()
+        account._set_session()
 
         # retrieves the session identifier (sid) for the currently
         # assigned session, this is going to be used in the next
@@ -979,6 +1219,17 @@ class AdminPart(
     def linked(self):
         return models.Settings.linked_apis()
 
+    def _counters(self):
+        adapter = self.get_adapter()
+        collection = adapter.collection("counters")
+        return collection
+
+    def _appier_extras_loader(self, module):
+        versions = []
+        if hasattr(module, "VERSION"):
+            versions.append(("AppierExtras", module.VERSION))
+        return versions
+
     def _netius_loader(self, module):
         versions = []
         if hasattr(module, "VERSION"):
@@ -1003,7 +1254,7 @@ class AdminPart(
         versions = []
         if hasattr(module, "VERSION"):
             version = ".".join([str(item) for item in module.VERSION])
-            versions.append(("redis-py", version))
+            versions.append(("RedisPy", version))
         return versions
 
     def _jinja2_loader(self, module):
@@ -1054,3 +1305,6 @@ class AdminPart(
             label = label,
             libs_label = libs_label
         )
+
+    def _to_meta(self, type):
+        return appier.Model._to_meta(type)
